import numpy as np


def get_uint_type_by_bitlength(bits):
    if bits == 8:
        return np.uint8
    elif bits == 16:
        return np.uint16
    elif bits == 32:
        return np.uint32
    elif bits == 64:
        return np.uint64
    else:
        raise NotImplementedError(f"mask_generator does not have this number of bits implemented: {bits}")



def mask_generator(bits=32, ones=9):
    """
    This function returns a mask of type unsigned int, which has the number of bits specified by the input variable bits 
    and has the number of ones specified by the input variable ones.

    Example:
        mask_generator(32,10) would return an unsigned integer whose binary representation would be 11111111110000000000000000000000

    FIXME: Right now we are not even considering little-big endian and these stuff.

    """
    mask_type = get_uint_type_by_bitlength(bits)
    # The approach to generate the mask is to generate a different single bit masks for each position that has a one
    # and then just merge them using logical_or
    elements = [single_bit_mask(position=i, bits=bits) for i in range(ones)]

    mask = mask_type(0.)
    for element in elements:
        mask = np.bitwise_or(mask, element)

    return mask_type(mask)


def single_bit_mask(position=0, bits=32):
    mask_type = get_uint_type_by_bitlength(bits)
    mask = mask_type(1)
    shift = mask_type(bits - position - 1)
    shifted = np.left_shift(mask, shift, casting="unsafe")
    return mask_type(shifted)


def apply_mask(array: np.ndarray, mask):

    bits = array.itemsize * 8
    int_type = get_uint_type_by_bitlength(bits)
    i_array = array.view(dtype=int_type)

    masked_array = np.bitwise_and(i_array, mask)
    
    assert len(masked_array.view(array.dtype)) == len(array)
    return masked_array.view(array.dtype)


def keep_significant_bits(array: np.ndarray, significant_bits=0, round_to_nearest=True):
    # For 32 bits:
    mask = mask_generator(ones=significant_bits)
    # round_to_nearest = True
    if round_to_nearest:
        # Get the mask for the first discarded value
        next_bit_mask = single_bit_mask(significant_bits)

        # Apply the mask
        next_bit_value = apply_mask(array, next_bit_mask)

        # Shift left
        next_bit_value = np.left_shift(next_bit_value.view(dtype=np.uint32), 1)

        # Apply or
        new_array = np.bitwise_or(array.view(dtype=np.uint32), next_bit_value).view(dtype=np.float32)
        if DEBUG: print("A:", binary_representation(array))
        if DEBUG: print("N:", binary_representation(new_array))

        # print(f" Flip probability: {1 - (sum(array == new_array)/array.size)}")
        array = new_array[:]

    if DEBUG: print("M:", [binary_representation(mask)] * array.size)

    pruned = apply_mask(array, mask)
    assert len(pruned) == len(array)

    if DEBUG: print("P:", binary_representation(pruned))
    return pruned


def binary_representation(A: np.ndarray):
    bits = A.itemsize * 8
    A = A.ravel()

    int_type = get_uint_type_by_bitlength(bits)
    # Create a integer view of the data to be usable with binary_repr
    A = A.view(dtype=int_type)
    
    def wrapper(v):
        return np.binary_repr(v, width=bits)

    wrapper_vectorized = np.vectorize(wrapper)

    try:
        if isinstance(A, np.ndarray):
            return wrapper_vectorized(A)
        return np.binary_repr(A, bits)
    except Exception as err:
        print(type(A))
        raise err


def bit_in_position(A, position):
    assert A.ndim == 1

    # The number of bits per value its equal to the array type size in bytes multiplied by 8
    byte_length = A.dtype.itemsize
    bits = byte_length * 8

    shift = bits - position - 1
    mask = single_bit_mask(position=position, bits=bits)
    x = np.bitwise_and(A.view(dtype=mask.dtype), mask)
    x = np.right_shift(x, shift)
    
    return np.uint8(x)


def bit_count(A, position):
    x = bit_in_position(A, position=position)
    return np.sum(x)


def bit_probabilities(A):
    # The number of bits per value its equal to the array type size in bytes multiplied by 8
    byte_length = A.dtype.itemsize
    bit_length = byte_length * 8
    _bits = range(bit_length)

    counted_bits = [bit_count(A, pos) for pos in _bits]
    return [c / (len(A)) for c in counted_bits]


def entropy_(p):
    s = 0.
    z = 0.
    for pi in p:
        if pi > z:
            s += pi * np.log(pi)
    return -s if s < 0 else s


def entropy__(p, base):
    return entropy_(p) / np.log(base)


def bitpattern_entropy_per_bit(A):
    probabilities = bit_probabilities(A)
    return [entropy__([p, 1-p], base=2) for p in probabilities]


def bitpattern_entropy(A):
    return new_entropy(A)


def new_entropy(array):
    unique, counts = np.unique(array, return_counts=True)
    probabilities = counts / array.size
    p = probabilities
    E = - np.sum(p * np.log2(p))
    return E


def bit_conditional_count(array):
    l_not = np.logical_not
    
    as_bool = array.astype(np.bool)
    rolled = np.roll(as_bool, 1)

    zero_zero = np.logical_and(l_not(as_bool), l_not(rolled))
    zero_one = np.logical_and(l_not(as_bool), rolled)
    one_zero = np.logical_and(as_bool, l_not(rolled))
    one_one = np.logical_and(as_bool, rolled)

    counter = np.zeros(shape=(2, 2))
    counter[0][0] = np.sum(zero_zero)
    counter[0][1] = np.sum(zero_one)
    counter[1][0] = np.sum(one_zero)
    counter[1][1] = np.sum(one_one)

    return counter


def bit_mutual_information(array):
    
    counter = bit_conditional_count(array)
    probabilities = counter / array.size
    p = probabilities

    p1 = np.sum(array) / array.size
    p0 = 1. - p1
    
    pc = np.zeros(shape=(2, 2))
    pc[0][0] = (probabilities[0][0] / p0) if p0 > 0. else 0
    pc[0][1] = (probabilities[0][1] / p0) if p0 > 0. else 0
    pc[1][0] = (probabilities[1][0] / p1) if p1 > 0. else 0
    pc[1][1] = (probabilities[1][1] / p1) if p1 > 0. else 0

    H0 = - ((pc[0][0] * np.log2(pc[0][0]) if pc[0][0] > 0 else 0) + (
        pc[0][1] * np.log2(pc[0][1]) if pc[0][1] > 0 else 0))
    H1 = - ((pc[1][0] * np.log2(pc[1][0]) if pc[1][0] > 0 else 0) + (
        pc[1][1] * np.log2(pc[1][1]) if pc[1][1] > 0 else 0))

    H = entropy__([p0, p1], 2)

    mutual_information = H - p0 * H0 - p1 * H1
    if mutual_information < 0:
        mutual_information = 0
    return mutual_information


def array_mutual_information(array):
    
    width = array.itemsize * 8

    mutual_information = []
    for pos in range(width):
        bits = bit_in_position(array, position=pos)
        mutual_information.append(bit_mutual_information(bits))


    mutual_information = np.array(mutual_information)
    return mutual_information
